#!/usr/bin/env python
#
# (c) 2011. Ivan Voras <ivoras@freebsd.org>
# Released under the BSDL
# This script will process an INDEX file and output a build list which
# can be parallelized.
#

import sys,os,re
from getopt import getopt

def usage():
    print "usage: %s -i index_file [-m max_qlen] [-v] | -h" % sys.argv[0]
    print
    print "\t-m\tMaximum parallel queue len"
    print "\t-v\tBe verbose - include comments in output"

index_file = None
max_qlen = 8
flag_verbose = False

opts, args = getopt(sys.argv[1:], "hi:m:v")
for o, a in opts:
    if o == '-i':
        index_file = a
    elif o == "-m":
        max_qlen = int(a)
    elif o == "-v":
        flag_verbose = True
    elif o == '-h':
        usage()
        sys.exit(0)

if index_file == None:
    usage()
    sys.exit(0)

def pname_split(name):
    i = name.rfind("-")
    return (name[0:i], name[i+1:])

index = {}

for line in file(index_file):
    ispec = line.split("|")
    name, ver = pname_split(ispec[0])

    index[name] = {
        "name"      : name,
        "ver"       : ver,
        "pname"     : ispec[0],
        "dir"       : ispec[1],
        "prefix"    : ispec[2],
        "descr"     : ispec[3],
        "ldescr"    : ispec[4],
        "maint"     : ispec[5],
        "categs"    : ispec[6].split(" "),
        "deps"      : [pname_split(x)[0] for x in ispec[7].split(" ") if x != ""],
        "done"      : False,
        "visited"   : False,
    }

flatindex = []
def dfs(port):
    global index, flatindex

    for dn in port["deps"]:
        if index[dn]["visited"]:
            continue
        index[dn]["visited"] = True
        dfs(index[dn])
        flatindex.append(dn)


for p in index:
    dfs(index[p])

qbatch = []
batch = []
for p in flatindex:
    for pd in index[p]["deps"]:
        if pd in batch:
            while len(batch) > 0:
                qbatch.append(batch[0:max_qlen])
                del batch[0:max_qlen]
            assert batch == []
    batch.append(p)
    index[p]["done"] = True

for p in index:
    if not index[p]["done"]:
        batch.append(p)

while len(batch) > 0:
    q = batch[0:max_qlen]
    del batch[0:max_qlen]
    qbatch.append(q)

def verify_strict_deps():
    """
    Verify that strictly, all deps of a port are satisfied before
    it is built.
    """
    global qbatch, index, flag_verbose
    bqueue = []

    for q in qbatch:
        for p in q:
            for pd in index[p]["deps"]:
                if not pd in bqueue:
                    raise "Uh-oh, dependancy %s of port %s not built before the port!" % (pd, p)
            bqueue.append(p)
    if flag_verbose:
        print "# Build order correct"

verify_strict_deps()

for batch in qbatch:
    print " ".join(batch)

if flag_verbose:
    print "# Maximum queue size:", max_qlen
    qallsize = 0
    for batch in qbatch:
        qallsize += len(batch)
    print "# Average batch size: %0.1f" % (float(qallsize) / len(qbatch))
